package main

import (
	"flag"
	"github.com/naleek/funmow"
	"log"
	"net"
)

const (
	FunMOWVersion       = "0.2"
	FunMOWDefaultDB     = "funmow.db"
	FunMOWDefaultListen = ":4201"
)

func main() {

	var seed bool
	var listen string
	var dbPath string

	flag.BoolVar(&seed, "seed", false, "Seed the world DB")
	flag.StringVar(&listen, "listen", FunMOWDefaultListen, "Listen Address:Port of game server.")
	flag.StringVar(&dbPath, "db", FunMOWDefaultDB, "Path to FunMOW database.")

	help := flag.Bool("h", false, "Show usage")

	flag.Parse()

	if *help {
		flag.PrintDefaults()
		return
	}

	log.Print("Starting FunMOW Version ", FunMOWVersion)

	log.Print("Using database ", FunMOWDefaultDB)
	db := funmow.NewDB(FunMOWDefaultDB)
	db.Open()

	defer db.Close()

	if seed {
		log.Print("Seeding database...")
		seedDB(db)
		return
	}

	log.Print("Listening for connections on ", listen)
	ln, err := net.Listen("tcp", listen)
	if err != nil {
		log.Fatal(err)
	}

	log.Print("Starting event distributor...")

	e := funmow.NewEventDistributor(db)
	go e.Run()

	log.Print("Waiting for connections...")

	sessionCounter := 0
	for {
		conn, _ := ln.Accept()
		c := funmow.NewClient(conn, sessionCounter, e, db)
		sessionCounter++
		go c.Run()
	}
}

func seedDB(db *funmow.DB) {

	f := funmow.NewObjectFactory(db)

	limbo := f.NewRawObject()
	limbo.Type = "room"
	limbo.Name = "Limbo"
	limbo.Description = "There's very little to see here."
	limbo.Owner = 1
	limbo.SetFlag("jumk_ok", true)
	limbo.ID = 0
	err := limbo.Commit()
	if err != nil {
		log.Fatal(err)
	}

	keelan, _ := db.CreatePlayer("keelan", "123", map[string]bool{"wizard": true})
	dan, _ := db.CreatePlayer("dan", "123", nil)

	outside := f.NewObject()
	outside.Type = "room"
	outside.Name = "outside"
	outside.Description = "The great outdoors. It's scary out here."

	outside.Commit()

	ketamine := f.NewObject()
	ketamine.Type = "thing"
	ketamine.Name = "ketamine"
	ketamine.Description = "You see a small pill bottle containing roughly 20 ketamine."

	ketamine.Commit()
	outside.Contains(&ketamine)

	yacht := f.NewObject()
	yacht.Type = "thing"
	yacht.Name = "gravy yacht"
	yacht.Description = "A gravy yacht. Like a boat but bigger. Duh."

	yacht.Commit()
	outside.Contains(&yacht)

	kitchen := f.NewObject()
	kitchen.Type = "room"
	kitchen.Name = "kitchen"
	kitchen.Description = "a lovely kitchen"

	kitchen.Commit()

	kettle := f.NewObject()
	kettle.Type = "thing"
	kettle.Name = "kettle"
	kettle.Description = "a lovely little kettle"

	kettle.Commit()
	kitchen.Contains(&kettle)

	bathroom := f.NewObject()
	bathroom.Type = "room"
	bathroom.Name = "bathroom"
	bathroom.Description = "a dirty bathroom"

	bathroom.Commit()

	toilet := f.NewObject()
	toilet.Type = "thing"
	toilet.Name = "toilet"
	toilet.Description = "a nasty toilet"

	toilet.Commit()
	bathroom.Contains(&toilet)

	door1 := f.NewObject()
	door1.Type = "exit"
	door1.Name = "<B>athroom;b;bathroom"
	door1.Description = "a door"
	door1.Next = bathroom.ID

	door1.Commit()

	door2 := f.NewObject()
	door2.Type = "exit"
	door2.Name = "<O>utside;o;outside"
	door2.Description = "a door"
	door2.Next = outside.ID

	door2.Commit()

	door3 := f.NewObject()
	door3.Type = "exit"
	door3.Name = "<K>itchen;k;kitchen"
	door3.Description = "a door"
	door3.Next = kitchen.ID

	door3.Commit()

	door4 := f.NewObject()
	door4.Type = "exit"
	door4.Name = "<H>ouse;h;house"
	door4.Description = "a door"
	door4.Next = kitchen.ID

	door4.Commit()

	kitchen.Contains(&door1)
	kitchen.Contains(&door2)
	bathroom.Contains(&door3)
	outside.Contains(&door4)

	db.Link(kitchen.ID, keelan, "player")

	db.Link(bathroom.ID, dan, "player")
}
